@@ -30,6 +30,15 @@
 #include <cogl/cogl-wayland-server.h>
 #include <meta/util.h>
 
+enum
+{
+  RESOURCE_DESTROYED,
+
+  LAST_SIGNAL
+};
+
+guint signals[LAST_SIGNAL];
+
 G_DEFINE_TYPE (MetaWaylandBuffer, meta_wayland_buffer, G_TYPE_OBJECT);
 
 static void
@@ -40,29 +49,10 @@ meta_wayland_buffer_destroy_handler (struct wl_listener *listener,
     wl_container_of (listener, buffer, destroy_listener);
 
   buffer->resource = NULL;
-  wl_signal_emit (&buffer->destroy_signal, buffer);
+  g_signal_emit (buffer, signals[RESOURCE_DESTROYED], 0);
   g_object_unref (buffer);
 }
 
-void
-meta_wayland_buffer_ref_use_count (MetaWaylandBuffer *buffer)
-{
-  g_warn_if_fail (buffer->resource);
-
-  buffer->use_count++;
-}
-
-void
-meta_wayland_buffer_unref_use_count (MetaWaylandBuffer *buffer)
-{
-  g_return_if_fail (buffer->use_count != 0);
-
-  buffer->use_count--;
-
-  if (buffer->use_count == 0 && buffer->resource)
-    wl_resource_queue_event (buffer->resource, WL_BUFFER_RELEASE);
-}
-
 MetaWaylandBuffer *
 meta_wayland_buffer_from_resource (struct wl_resource *resource)
 {
@@ -82,7 +72,6 @@ meta_wayland_buffer_from_resource (struct wl_resource *resource)
       buffer = g_object_new (META_TYPE_WAYLAND_BUFFER, NULL);
 
       buffer->resource = resource;
-      wl_signal_init (&buffer->destroy_signal);
       buffer->destroy_listener.notify = meta_wayland_buffer_destroy_handler;
       wl_resource_add_destroy_listener (resource, &buffer->destroy_listener);
     }
@@ -98,7 +87,6 @@ meta_wayland_buffer_ensure_texture (MetaWaylandBuffer *buffer)
   CoglTexture *texture;
   struct wl_shm_buffer *shm_buffer;
 
-  g_return_val_if_fail (buffer->use_count != 0, NULL);
   g_return_val_if_fail (buffer->resource, NULL);
 
   if (buffer->texture)
@@ -134,8 +122,6 @@ meta_wayland_buffer_process_damage (MetaWaylandBuffer *buffer,
 {
   struct wl_shm_buffer *shm_buffer;
 
-  g_return_if_fail (buffer->use_count != 0);
-
   shm_buffer = wl_shm_buffer_get (buffer->resource);
 
   if (shm_buffer)
@@ -181,4 +167,11 @@ meta_wayland_buffer_class_init (MetaWaylandBufferClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
   object_class->finalize = meta_wayland_buffer_finalize;
+
+  signals[RESOURCE_DESTROYED] = g_signal_new ("resource-destroyed",
+                                              G_TYPE_FROM_CLASS (object_class),
+                                              G_SIGNAL_RUN_LAST,
+                                              0,
+                                              NULL, NULL, NULL,
+                                              G_TYPE_NONE, 0);
 }
